@@ -5,10 +5,10 @@
 #
 # File        : jiten/cli.py
 # Maintainer  : Felix C. Stegerman <flx@obfusk.net>
-# Date        : 2020-08-31
+# Date        : 2021-01-22
 #
-# Copyright   : Copyright (C) 2020  Felix C. Stegerman
-# Version     : v0.3.3
+# Copyright   : Copyright (C) 2021  Felix C. Stegerman
+# Version     : v0.3.5
 # License     : AGPLv3+
 #
 # --                                                            ; }}}1
@@ -40,7 +40,7 @@ CLI
 <BLANKLINE>
 
 >>> run("-v jmdict -m1 -w cat")
-DB v8 up to date.
+DB v11 up to date.
 query: +w cat
 <BLANKLINE>
 猫
@@ -56,12 +56,12 @@ query: +w cat
 --> noun (common) (futsuumeishi) | abbreviation | word usually written using
     kana alone | colloquialism
 see 猫車 | 猫火鉢
-seq# 1467640, freq# 2201, prio; 1
+seq# 1467640, freq# 2201, jlpt N5, prio; 1
 <BLANKLINE>
 <BLANKLINE>
 
 >>> run("-v jmdict -m1 -w kat -l dut")
-DB v8 up to date.
+DB v11 up to date.
 query: +w kat
 <BLANKLINE>
 猫
@@ -74,12 +74,12 @@ query: +w kat
 = kruiwagen
 --> noun (common) (futsuumeishi)
 see 猫車 | 猫火鉢
-seq# 1467640, freq# 2201, prio; 1
+seq# 1467640, freq# 2201, jlpt N5, prio; 1
 <BLANKLINE>
 <BLANKLINE>
 
 >>> run("-v jmdict -m1 -w idiot")
-DB v8 up to date.
+DB v11 up to date.
 query: +w idiot
 <BLANKLINE>
 馬鹿 | 莫迦 | 破家 | 馬稼
@@ -96,12 +96,12 @@ query: +w idiot
     abbreviation
 ~~> ateji (phonetic) reading
 see 馬鹿貝
-seq# 1601260, freq# 2472, prio; 1
+seq# 1601260, freq# 2472, jlpt N3, prio; 1
 <BLANKLINE>
 <BLANKLINE>
 
 >>> run("-v jmdict -m1 -e 誤魔化す")
-DB v8 up to date.
+DB v11 up to date.
 query: += 誤魔化す
 <BLANKLINE>
 誤魔化す | 誤摩化す | 胡麻化す | 誤魔かす | 胡魔化す
@@ -112,15 +112,15 @@ query: += 誤魔化す
   tamper | to juggle | to manipulate
 = to dodge | to beg the question (issue, difficulties)
 = to varnish over | to gloss over
---> Godan verb with `su' ending | transitive verb | word usually written using
+--> Godan verb with 'su' ending | transitive verb | word usually written using
     kana alone
 ~~> ateji (phonetic) reading
-seq# 1271480, freq# 10495, prio; 1
+seq# 1271480, freq# 10495, jlpt N1, prio; 1
 <BLANKLINE>
 <BLANKLINE>
 
 >>> run("-v jmdict -m1 -w まる")
-DB v8 up to date.
+DB v11 up to date.
 query: +w まる
 <BLANKLINE>
 丸 | 円
@@ -138,17 +138,17 @@ query: +w まる
 --> noun (common) (futsuumeishi) | noun, used as a prefix | suffix | slang |
     esp. 丸
 see ○ | まる | スッポン | 麻呂
-seq# 1216250, freq# 63, prio; 1
+seq# 1216250, freq# 63, jlpt N3, prio; 1
 <BLANKLINE>
 <BLANKLINE>
 
 >>> run("-v jmdict -m1 -w cat --verb")
-DB v8 up to date.
+DB v11 up to date.
 query: +w cat
 <BLANKLINE>
 逆撫で | 逆なで
 さかなで | ぎゃくなで
-[no pitch data]
+さꜛかなで
 [eng]
 = rubbing the wrong way (e.g. a cat) | irritating
 --> noun (common) (futsuumeishi) | noun or participle which takes the aux. verb
@@ -159,7 +159,7 @@ seq# 1227180, freq# 30500; 1
 <BLANKLINE>
 
 >>> run("-v jmdict -m1 -w みる --noun")
-DB v8 up to date.
+DB v11 up to date.
 query: +w みる
 <BLANKLINE>
 海松 | 水松
@@ -175,7 +175,7 @@ seq# 1772790, freq# 75; 1
 <BLANKLINE>
 
 >>> run("-v jmdict -m1 -w みる --noun --prio")
-DB v8 up to date.
+DB v11 up to date.
 query: +w みる
 <BLANKLINE>
 <BLANKLINE>
@@ -193,7 +193,7 @@ query: +w みる
 <BLANKLINE>
 
 >>> run("-v kanji -m1 -e cat")
-DB v8 up to date.
+DB v11 up to date.
 query: += cat
 <BLANKLINE>
 猫
@@ -202,9 +202,9 @@ query: += cat
 [no name readings]
 = cat
 --> 猫 【ねこ】 | 子猫 【こねこ】 | 猫背 【ねこぜ】 |
-    山猫 【やまねこ】 | 猫舌 【ねこじた】 | 猫 【ねこま】 |
-    野良猫 【のらねこ】 | 猫撫で声 【ねこなでごえ】 | 猫又 【ねこまた】 |
-    黒猫 【くろねこ】 | 飼い猫 【かいねこ】 | 招き猫 【まねきねこ】 |
+    山猫 【やまねこ】 | 猫舌 【ねこじた】 | 野良猫 【のらねこ】 |
+    黒猫 【くろねこ】 | 飼い猫 【かいねこ】 | 猫 【ねこま】 |
+    猫撫で声 【ねこなでごえ】 | 猫又 【ねこまた】 | 招き猫 【まねきねこ】 |
     猫足 【ねこあし】 | 斑猫 【はんみょう】 | 猫かぶり 【ねこかぶり】 |
     海猫 【うみねこ】 | 猫柳 【ねこやなぎ】 | 猫板 【ねこいた】 |
     トラ猫 【トラねこ】 | 猫可愛がり 【ねこかわいがり】 |
@@ -213,12 +213,12 @@ query: += cat
 部: 犬 (94) ⺨ 犯 田 艸 艹 艾 苗
 variants: 貓
 0x732b CJK UNIFIED IDEOGRAPH-732B; 1
-11 strokes, level 常用, freq# 1702, old jlpt 2, jlpt 3, skip 1-3-8
+11 strokes, level 常用, freq# 1702, old jlpt N2, jlpt N3, skip 1-3-8
 <BLANKLINE>
 <BLANKLINE>
 
 >>> run("-v kanji -m1 -w 日")
-DB v8 up to date.
+DB v11 up to date.
 query: +w 日
 <BLANKLINE>
 日
@@ -236,21 +236,15 @@ query: +w 日
     訪日 【ほうにち】 | 三十日 【さんじゅうにち】 | 三十日 【みそか】
 部: 日 (72)
 0x65e5 CJK UNIFIED IDEOGRAPH-65E5; 1
-4 strokes, level 常用1, freq# 1, old jlpt 4, jlpt 5, skip 3-3-1
+4 strokes, level 常用1, freq# 1, old jlpt N4, jlpt N5, skip 3-3-1
 <BLANKLINE>
 <BLANKLINE>
 
 """                                                             # }}}1
 
-import os, os.path as _pth, subprocess as _subp, sys
+import os, sys
 
-_cwd = _pth.dirname(__file__)
-if _pth.exists(_pth.join(_pth.dirname(_cwd), ".git")):
-  __version__ = _subp.check_output("git describe --always",
-                  shell = True, cwd = _cwd) \
-                .decode().strip().replace("v", "", 1)
-else:
-  __version__ = "0.3.3"
+from .version import __version__, py_version
 name = "jiten"
 
 import click
@@ -261,24 +255,56 @@ from . import misc      as M
 from . import pitch     as P
 from . import sentences as S
 
-def setup_db(verbose):
+from .kana import with_romaji
+from .misc import SERVER
+
+HOST, PORT    = "localhost", 5000
+MODS          = [K, P, S, J] # J last!
+ANDROID_PRIV  = os.environ.get("ANDROID_PRIVATE") or None
+
+def setup_db(verbose, dl = False):
   msg = "up to date"
   if not J.up2date():
-    msg = "set up"
-    K.setup()
-    P.setup()
-    S.setup()
-    J.setup() # last
+    if dl or missing_data():
+      msg = "downloaded"
+      download_dbs()
+    else:
+      msg = "set up"
+      for m in MODS: m.setup()
   if verbose:
     click.secho("DB v{} {}.".format(J.DBVERSION, msg), fg = "green")
 
+def missing_data():
+  for m in MODS:
+    for f in m.DATA_FILES[1:]:
+      if not os.path.exists(f): return True
+  return False
+
+def download_dbs():
+  url = lambda x: "{}/_db/v{}/{}".format(SERVER, J.DBVERSION, x)
+  for file in [ m.DATA_FILES[0] for m in MODS ]:
+    fname = os.path.basename(file)
+    base  = os.path.splitext(fname)[0]
+    f     = os.path.join(ANDROID_PRIV, fname) if ANDROID_PRIV else file
+    M.download_file(url(base), f, M.DB_SHA512SUMS[J.DBVERSION][base])
+  if ANDROID_PRIV: android_link_dbs()
+
+def android_link_dbs():
+  for file in [ m.DATA_FILES[0] for m in MODS ]:
+    f = os.path.join(ANDROID_PRIV, os.path.basename(file))
+    if os.path.exists(f) and not os.path.islink(file):
+      os.symlink(f, file)
+
 @click.group(help = """
-  jiten - japanese cli&web dictionary based on jmdict/kanjidic
+  jiten - japanese android/cli/web dictionary based on jmdict/kanjidic
 """)
 @click.option("-v", "--verbose", is_flag = True, help = "Be verbose.")
 @click.option("-c", "--colour/--no-colour", is_flag = True,
               default = None, help = "Use terminal colours.")
-@click.version_option(__version__)
+@click.version_option(
+  "{} [{}] [DB v{}]".format(__version__, py_version, J.DBVERSION),
+  message = "%(prog)s 「辞典」 %(version)s"
+)
 @click.pass_context
 def cli(ctx, colour, **kw):
   if colour is not None: ctx.color = colour
@@ -301,7 +327,10 @@ def cli(ctx, colour, **kw):
 @click.option("--noun", is_flag = True, help = "Select nouns.")
 @click.option("--verb", is_flag = True, help = "Select verbs.")
 @click.option("--prio", is_flag = True, help = "Select priority entries.")
-@click.argument("query", required = False, metavar = "REGEX")
+@click.option("-n", "--jlpt", type = M.JLPT_LEVEL,
+              help = "Select entries by JLPT level(s); e.g. 1 or 3-5.")
+@click.option("--romaji", is_flag = True, help = "Show romaji.")
+@click.argument("query", required = False)
 @click.pass_context
 def jmdict(ctx, query, **kw):
   setup_db(ctx.obj["verbose"])
@@ -314,10 +343,12 @@ def jmdict(ctx, query, **kw):
       if not q: break
       click.echo_via_pager(jmdict_search(q, **ctx.obj))
 
-def jmdict_search(q, verbose, word, exact, fstwd, langs, **kw): # {{{1
+def jmdict_search(q, verbose, word, exact, fstwd, langs, romaji,
+                  **kw):                                        # {{{1
   langs = [ l for ls in langs for l in ls.split(",") ]
-  q     = M.process_query(q, word, exact, fstwd)
+  q     = M.process_query(q, word, exact, fstwd, True)
   w     = click.get_terminal_size()[0]
+  f     = with_romaji if romaji else lambda x: x
   if verbose:
     yield "query: " + click.style(q, fg = "bright_red") + "\n\n"
   for i, (e, rank) in enumerate(J.search(q, langs = langs, **kw)):
@@ -325,11 +356,14 @@ def jmdict_search(q, verbose, word, exact, fstwd, langs, **kw): # {{{1
       click.style(k.elem, fg = "bright_yellow") for k in e.kanji
     ) or "[no kanji]") + "\n"
     yield (" | ".join(
-      click.style(r.elem, fg = "bright_green") for r in e.reading
+      click.style(f(r.elem), fg = "bright_green") for r in e.reading
     ) or "[no readings]") + "\n"
-    yield (" | ".join(
-      click.style(p, fg = "cyan") for p in e.pitch()
-    ) or "[no pitch data]") + "\n"
+    if P.have_pitch():
+      yield (" | ".join(
+        click.style(f(p), fg = "cyan") for p in e.pitch()
+      ) or "[no pitch data]") + "\n"
+    else:
+      yield "[pitch data unavailable]\n"
     gloss, info = e.gloss_pos_info(langs)
     for l in langs:
       yield click.style("[" + l + "]", fg = "blue") + "\n"
@@ -345,6 +379,8 @@ def jmdict_search(q, verbose, word, exact, fstwd, langs, **kw): # {{{1
       yield  ("seq# " + click.style(str(e.seq), fg = "blue")
         + (", freq# " + click.style(str(rank ), fg = "cyan")
                     if rank else "")
+        + (", jlpt " + click.style("N"+str(e.jlpt), fg = "yellow")
+                    if e.jlpt else "")
         + (", prio" if e.isprio() else "") + "; " + str(i+1) + "\n")
     yield "\n"
                                                                 # }}}1
@@ -367,7 +403,7 @@ def indent_and_wrap_jap(w, xs, pre, fg):
   w = len(pre) + n + (n // k * 2)   # 2 extra chars per word per line
   return indent_and_wrap(w, xs, pre, fg).replace("_【", " 【")
 
-@cli.command(help = "Search Kanji.")
+@cli.command(help = "Search kanji.")
 @click.option("-w", "--word", is_flag = True,
               help = "Match whole word (same as \\b...\\b).")
 @click.option("-1", "--1stword", "--first-word", "fstwd", is_flag = True,
@@ -376,7 +412,8 @@ def indent_and_wrap_jap(w, xs, pre, fg):
               help = "Match exactly (same as ^...$).")
 @click.option("-m", "--max", "max_results", default = None,
               type = click.INT, help = "Maximum number of results.")
-@click.argument("query", required = False, metavar = "REGEX")
+@click.option("--romaji", is_flag = True, help = "Show romaji.")
+@click.argument("query", required = False)
 @click.pass_context
 def kanji(ctx, query, **kw):
   setup_db(ctx.obj["verbose"])
@@ -389,21 +426,23 @@ def kanji(ctx, query, **kw):
       if not q: break
       click.echo_via_pager(kanji_search(q, **ctx.obj))
 
-def kanji_search(q, verbose, word, exact, fstwd, max_results):  # {{{1
-  q = M.process_query(q, word, exact, fstwd)
+def kanji_search(q, verbose, word, exact, fstwd, max_results,
+                 romaji):                                       # {{{1
+  q = M.process_query(q, word, exact, fstwd, True)
   w = click.get_terminal_size()[0]
+  f = with_romaji if romaji else lambda x: x
   if verbose:
     yield "query: " + click.style(q, fg = "bright_red") + "\n\n"
   for i, e in enumerate(K.search(q, max_results)):
     yield e.char + "\n"
     yield (" | ".join(
-      click.style(r, fg = "bright_yellow") for r in e.on
+      click.style(f(r), fg = "bright_yellow") for r in e.on
     ) or "[no on readings]") + "\n"
     yield (" | ".join(
-      click.style(r, fg = "bright_green") for r in e.kun
+      click.style(f(r), fg = "bright_green") for r in e.kun
     ) or "[no kun readings]") + "\n"
     yield (" | ".join(
-      click.style(r, fg = "cyan") for r in e.nanori
+      click.style(f(r), fg = "cyan") for r in e.nanori
     ) or "[no name readings]") + "\n"
     if e.meaning:
       yield click.style("= ", fg = "magenta") + \
@@ -431,23 +470,23 @@ def kanji_search(q, verbose, word, exact, fstwd, max_results):  # {{{1
            if e.level else "")
         + (", freq# " + click.style(str(e.freq), fg = "magenta")
            if e.freq else "")
-        + (", old jlpt " + click.style(str(e.jlpt), fg = "blue")
+        + (", old jlpt " + click.style("N"+str(e.jlpt), fg = "blue")
            if e.jlpt else "")
-        + (", jlpt " + click.style(str(e.new_jlpt()), fg = "cyan")
-           if e.new_jlpt() else "")
+        + (", jlpt " + click.style("N"+str(e.new_jlpt), fg = "cyan")
+           if e.new_jlpt else "")
         + (", skip " + click.style(e.skip, fg = "yellow")
            if e.skip else "") + "\n")
     yield "\n"
                                                                 # }}}1
 
-@cli.command(help = "Search Example Sentences.")
+@cli.command(help = "Search Tatoeba example sentences.")
 @click.option("-l", "--lang", "langs", multiple = True,
               default = [], metavar = "LANG",
               envvar = name.upper() + "_LANGS",
               help = "Filter language(s) ("+", ".join(S.LANGS)+").")
 @click.option("-m", "--max", "max_results", default = None,
               type = click.INT, help = "Maximum number of results.")
-@click.argument("query", required = False, metavar = "STRING")
+@click.argument("query", required = False)
 @click.pass_context
 def sentences(ctx, query, **kw):
   setup_db(ctx.obj["verbose"])
@@ -477,18 +516,41 @@ def sentence_search(q, verbose, langs, max_results):            # {{{1
                                                                 # }}}1
 
 @cli.command(help = "Serve the web interface.")
-@click.option("-h", "--host", default = "localhost", metavar = "HOST")
-@click.option("-p", "--port", default = 5000, metavar = "PORT", type = click.INT)
+@click.option("-h", "--host", default = HOST, metavar = "HOST",
+              help = "Host to listen on.", show_default = True)
+@click.option("-p", "--port", default = PORT, metavar = "PORT",
+              help = "Port to listen on.", show_default = True,
+              type = click.INT)
 @click.pass_context
 def serve(ctx, host, port):
-  setup_db(ctx.obj["verbose"])
-  from .app import app
-  app.run(host = host, port = port, load_dotenv = False)
+  serve_app(host, port, ctx.obj["verbose"])
 
-@cli.command(help = "Create sqlite databases from XML files.")
-def setup():
+def serve_app(host = HOST, port = PORT, verbose = True, **opts):
+  from .app import app
+  if ANDROID_PRIV:
+    android_link_dbs()
+    app.config["DBS_UP2DATE"]   = J.up2date()
+    app.config["DOWNLOAD_DBS"]  = download_dbs
+  else:
+    setup_db(verbose)
+  app.run(host = host, port = port, load_dotenv = False, **opts)
+
+@cli.command(help = """
+  WebView GUI.  Wraps the web interface.  Requires pywebview.
+""")
+@click.argument("link", required = False)
+@click.pass_context
+def gui(ctx, link):
+  setup_db(ctx.obj["verbose"])
+  from .gui import start
+  start(link)
+
+@cli.command(help = "Build (or download) sqlite databases.")
+@click.option("--download", is_flag = True,
+              help = "Always download DBs, never build them.")
+def setup(download):
   click.echo("Creating databases...")
-  setup_db(True)
+  setup_db(True, download)
 
 @cli.command("_doctest", hidden = True)
 @click.pass_context
