@@ -5,10 +5,10 @@
 #
 # File        : jiten/jmdict.py
 # Maintainer  : Felix C. Stegerman <flx@obfusk.net>
-# Date        : 2020-08-31
+# Date        : 2021-01-22
 #
-# Copyright   : Copyright (C) 2020  Felix C. Stegerman
-# Version     : v0.3.2
+# Copyright   : Copyright (C) 2021  Felix C. Stegerman
+# Version     : v0.3.5
 # License     : AGPLv3+
 #
 # --                                                            ; }}}1
@@ -19,29 +19,57 @@ r"""
 JMDict.
 
 >>> DBVERSION
-8
+11
 
 >>> jmdict = parse_jmdict()
 >>> len(jmdict)
-188785
+190722
 
 >>> print(jmdict[-1].sense[0].gloss[0])
-Japanese-Multilingual Dictionary Project - Creation Date: 2020-07-31
+Japanese-Multilingual Dictionary Project - Creation Date: 2021-01-19
 
 >>> len([ x for x in jmdict if x.isprio() ])
-22437
+22443
 >>> len([ x for x in jmdict if x.isprio() and list(x.pitch()) ])
-14714
+17415
+
+>>> len([ x for x in jmdict if x.jlpt == 1 ])
+3035
+>>> len([ x for x in jmdict if x.jlpt == 2 ])
+1725
+>>> len([ x for x in jmdict if x.jlpt == 3 ])
+1657
+>>> len([ x for x in jmdict if x.jlpt == 4 ])
+633
+>>> len([ x for x in jmdict if x.jlpt == 5 ])
+679
+
+>>> len([ x for x in jmdict if x.jlpt == 1 and x.isprio() ])
+2628
+>>> len([ x for x in jmdict if x.jlpt == 2 and x.isprio() ])
+1647
+>>> len([ x for x in jmdict if x.jlpt == 3 and x.isprio() ])
+1615
+>>> len([ x for x in jmdict if x.jlpt == 4 and x.isprio() ])
+615
+>>> len([ x for x in jmdict if x.jlpt == 5 and x.isprio() ])
+657
 
 >>> baka = [ x for x in jmdict if any( r.elem == "ばか" for r in x.reading ) ][0]
 >>> baka.seq
 1601260
 >>> baka.definition()
 ('ばか', 'バカ', '馬鹿', '莫迦', '破家', '馬稼')
+>>> tuple(baka.pitch())
+('ばꜜか',)
 >>> baka.usually_kana()
 True
+>>> baka.jlpt
+3
 >>> baka.isprio()
 True
+>>> baka.prio_level()
+11
 >>> baka.isnoun()
 True
 >>> baka.isverb()
@@ -90,10 +118,16 @@ ateji (phonetic) reading
 1578850
 >>> iku.definition()
 ('いく', 'ゆく', '行く', '逝く', '往く')
+>>> tuple(iku.pitch())
+('いꜛく', 'ゆꜛく')
 >>> iku.usually_kana()
 True
+>>> iku.jlpt
+5
 >>> iku.isprio()
 True
+>>> iku.prio_level()
+10
 >>> iku.isnoun()
 False
 >>> iku.isverb()
@@ -155,6 +189,12 @@ word containing out-dated kanji
 くる
 旨く行く
 
+>>> from .kana import kana2romaji
+>>> len([ kana2romaji(r.elem) for e in jmdict for r in e.reading ])
+228225
+>>> len([ kana2romaji(p) for e in jmdict for p in e.pitch() ])
+88402
+
 """                                                             # }}}1
 
 import gzip, os, re, sys
@@ -169,19 +209,24 @@ from . import misc  as M
 from . import pitch as P
 from .sql import sqlite_do, load_pcre_extension
 
-DBVERSION     = 8 # NB: update this when data/schema changes
-SQLITE_FILE   = M.resource_path("res/jmdict.sqlite3")
-JMDICT_FILE   = M.resource_path("res/jmdict/jmdict.xml.gz")
+DBVERSION       = 11 # NB: update this when data/schema changes
+SQLITE_FILE     = M.resource_path("res/jmdict.sqlite3")
+JMDICT_FILE     = M.resource_path("res/jmdict/jmdict.xml.gz")
+JLPT_FILE_BASE  = M.resource_path("res/jlpt/N")
+DATA_FILES      = (SQLITE_FILE, JMDICT_FILE)
 
-MAXSEQ        = 10000000
-PRIO          = "news1 ichi1 spec1 spec2 gai1".split()
-USUKANA       = "word usually written using kana alone"
-LANGS         = "eng dut ger".split()
+MAXSEQ          = 10000000
+PRIO            = dict(news1 = 10, news2 = 1, ichi1 = 10, ichi2 = 1,
+                       spec1 = 10, spec2 = 5, gai1  = 10, gai2  = 1)
+MINPRIO         = 5
+USUKANA         = "word usually written using kana alone"
+LANGS           = "eng dut ger".split()
+JLPTKK, JLPTUK  = JLPTKANA = "[katakana]", "[usukana]"
 
-Entry         = namedtuple("Entry"  , """seq kanji reading sense""".split())
-Kanji         = namedtuple("Kanji"  , """elem chars info prio""".split())
-Reading       = namedtuple("Reading", """elem restr info prio""".split())
-Sense         = namedtuple("Sense"  , """pos lang gloss info xref""".split())
+Entry   = namedtuple("Entry"  , """seq jlpt kanji reading sense""".split())
+Kanji   = namedtuple("Kanji"  , """elem chars info prio""".split())
+Reading = namedtuple("Reading", """elem restr info prio""".split())
+Sense   = namedtuple("Sense"  , """pos lang gloss info xref""".split())
 
 # TODO
 # * include all readings if usually_kana? or only frequent ones?
@@ -190,8 +235,7 @@ def definition(e):
   xs    = r + k if e.usually_kana() else k or r
   return tuple(M.uniq( x.elem for x in xs ))
 
-def words(e):
-  return frozenset( x.elem for x in M.flatten([e.kanji, e.reading]) )
+def words(e): return frozenset( x.elem for x in e.kanji + e.reading )
 
 def meanings(e, *a):
   l = a or [LANGS[0]]
@@ -209,10 +253,15 @@ def _freq(e): return sum( F.freq.get(w, 0) for w in e.definition() )
 # TODO: load from DB
 def _rank(e): return min( F.rank(w) for w in e.definition() )
 
-def isprio(e):
-  return any( x.prio for x in M.flatten([e.kanji, e.reading]) )
+def isprio(e):      return _isprio(e.kanji + e.reading)
+def _isprio(xs):    return (_prio_level(xs) or 0) >= MINPRIO
+def prio_level(e):  return _prio_level(e.kanji + e.reading)
+def _prio_level(xs):
+  ps = [ x.prio for x in xs if x.prio ]
+  return max(ps) if ps else None
 
-def usually_kana(e): return any( USUKANA in s.info for s in e.sense )
+def usually_kana(e): return _usukana(e.sense)
+def _usukana(sense): return any( USUKANA in s.info for s in sense )
 
 def gloss_pos_info(e, langs):
   gloss, pos, info = {}, [], []
@@ -242,6 +291,43 @@ def pitch(e):
     p = P.get_pitch(r.elem, ks)
     if p: yield p
 
+# TODO
+def jlpt_level(kanji, reading, usukana):                        # {{{1
+  kana, prio  = not kanji or usukana, _isprio(kanji + reading)
+  ls, ka      = set(), set( k.elem for k in kanji )
+  for k in ka:
+    jlpt = JLPT.get(k)
+    if not jlpt: continue
+    for r in reading:
+      ls.update( n for x, n in jlpt if r.elem == x )
+  for r in reading:
+    if r.elem in JLPT_COMMON:
+      if (ka and not ka & JLPT_CO_KA) or \
+         (not ka and r.elem not in JLPT_CO_NK): continue
+    jlpt = JLPT.get(r.elem)
+    if not jlpt: continue
+    for x, n in jlpt:
+      assert x in JLPTKANA
+      if x == JLPTUK and not kana:
+        if ka & JLPT_BLACK: continue
+        if not ka & JLPT_WHITE and not prio: continue
+      ls.add(n)
+  return max(ls) if ls else None
+                                                                # }}}1
+
+JLPT_BLACK  = set(""" 滑降 保母 """.split())
+JLPT_WHITE  = set(""" お先に お待たせしました お邪魔します 一昨昨日
+                      明々後日 """.split())
+JLPT_CO_KA  = set(""" 彼れ 一杯 斯う 此処 事 此の 先 為る 然う 其処
+                      其の 同 何の 成る 白 又 良く """.split())
+JLPT_CO_NK  = set(""" さん じゃ そう ちゃん と どうか なんか ね はい
+                  """.split())
+JLPT_COMMON = set(""" あ あした あと あれ いく いち いっぱい えい こう
+                      ここ こと この さっき さん し じゃ すり する
+                      せっけん そう そこ その たて ちゃん つき と どう
+                      どうか どの なる なんか ね はい はく ほんとう
+                      また よい よく """.split())               # TODO
+
 Entry.definition      = definition
 Entry.words           = words
 Entry.meanings        = meanings
@@ -251,6 +337,7 @@ Entry.chars           = chars
 Entry._freq           = _freq
 Entry._rank           = _rank
 Entry.isprio          = isprio
+Entry.prio_level      = prio_level
 Entry.usually_kana    = usually_kana
 Entry.gloss_pos_info  = gloss_pos_info
 Entry.isnoun          = isnoun
@@ -268,11 +355,14 @@ Sense.isintrans = lambda s: "intransitive verb" in s.pos
 Sense.istransitive    = Sense.istrans
 Sense.isintransitive  = Sense.isintrans
 
-def _isprio_k(e):
-  return any( x.text.strip() in PRIO for x in e.findall("ke_pri") )
+def _prio_l(tags):
+  return sum( PRIO.get(t, 0) for t in set(tags) ) # or None   # FIXME
 
-def _isprio_r(e):
-  return any( x.text.strip() in PRIO for x in e.findall("re_pri") )
+def _prio_k(e):
+  return _prio_l( x.text.strip() for x in e.findall("ke_pri") )
+
+def _prio_r(e):
+  return _prio_l( x.text.strip() for x in e.findall("re_pri") )
 
 def _kanji_chars(s): return frozenset( c for c in s if M.iskanji(c) )
 
@@ -294,7 +384,7 @@ def parse_jmdict(file = JMDICT_FILE):                           # {{{1
           keb   = ke.find("keb").text.strip()   # word/phrase w/ kanji
           info  = tuple( x.text.strip() for x in ke.findall("ke_inf") )
           assert all( "\n" not in x and "" not in x for x in info )
-          kanji.append(Kanji(keb, _kanji_chars(keb), info, _isprio_k(ke)))
+          kanji.append(Kanji(keb, _kanji_chars(keb), info, _prio_k(ke)))
         for re in e.findall("r_ele"):           # 1+ reading elem
           reb   = re.find("reb").text.strip()   # reading elem
           restr = tuple( x.text.strip() for x in re.findall("re_restr") )
@@ -302,15 +392,16 @@ def parse_jmdict(file = JMDICT_FILE):                           # {{{1
           info  = tuple( x.text.strip() for x in re.findall("re_inf") )
           assert all( "\n" not in x and "" not in x for xs in
                       [restr, info] for x in xs )
-          reading.append(Reading(reb, restr, info, _isprio_r(re)))
+          reading.append(Reading(reb, restr, info, _prio_r(re)))
         for se in e.findall("sense"):           # 1+ sense elem
           pos   = tuple( x.text.strip() for x in se.findall("pos") ) or pos
                   # part of speech, applies to following senses too
           lang, gloss = None, []
           for x in se.findall("gloss"):
-            if x.get(alang) in LANGS and x.text:
-              assert lang is None or lang == x.get(alang)
-              lang = x.get(alang)
+            l = x.get(alang, "eng")
+            if l in LANGS and x.text:
+              assert lang is None or lang == l
+              lang = l
               gloss.append(x.text.strip())
           if lang is None: continue
           s_inf = tuple( x.text.strip() for x in se.findall("s_inf") )
@@ -322,24 +413,67 @@ def parse_jmdict(file = JMDICT_FILE):                           # {{{1
           assert all( "\n" not in x and "" not in x for xs in
                       [pos, gloss, s_inf, misc, xref] for x in xs )
           sense.append(Sense(pos, lang, tuple(gloss), s_inf + misc, xref))
-        data.append(Entry(seq, *( tuple(x) for x in [kanji, reading, sense] )))
+        krs   = ( tuple(x) for x in [kanji, reading, sense] )
+        jlpt  = jlpt_level(kanji, reading, _usukana(sense))
+        data.append(Entry(seq, jlpt, *krs))
       return data
                                                                 # }}}1
 
+def load_jlpt(base = JLPT_FILE_BASE):                           # {{{1
+  skip        = set("×|Ͼ立|あげる (=やる)|より、ほう".split("|"))
+  kata        = "ローマじ ジェットき けしゴム".split()
+  repl        = { "あたたか(い)": "あたたかい" }
+  data, seen  = {}, set()
+  for level in "12345":
+    with open(base + level + "-vocab-hiragana") as f:
+      for line in f:
+        word, read = line.rstrip("\n").split("\t")
+        if word in skip: continue
+        if read.endswith("・する"): read = read[:-3]
+        words = word.replace("  ", "/").split("/")
+        reads = read.replace("  ", "/").replace(" / ", "/").split("/")
+        reads = [ repl.get(r, r) for r in reads ]
+        reads = [ r for r in reads if r in kata or M.ishiragana(r) ]
+        if not reads: continue
+        assert words and all( M.isokjap(w) for w in words )
+        assert not (len(words) > 1 and len(reads) > 1)
+        for w in words:
+          for r in reads:
+            assert w != r or M.iskana(w)
+            if w == r: r = JLPTUK
+            else: seen.add(w)
+            data.setdefault(w, []).append((r, int(level)))
+  for level in "12345":
+    with open(base + level + "-vocab-eng") as f:
+      for line in f:
+        words = line.rstrip("\n").replace("  ", "/").split("/")
+        words = [ w.strip() for w in words if w ]
+        if skip & set(words): continue
+        assert words and all( M.isokjap(w) for w in words )
+        for w in words:
+          if w not in seen:
+            assert M.iskana(w)
+            what = JLPTKK if M.iskatakana(w) else JLPTUK
+            data.setdefault(w, []).append((what, int(level)))
+  return data
+                                                                # }}}1
+
+JLPT = load_jlpt()                                              # TODO
+
 # NB: kanji/reading/sense are retrieved in insertion (i.e. rowid) order!
 def jmdict2sqldb(data, file = SQLITE_FILE):                     # {{{1
   with sqlite_do(file) as c:
     c.executescript(JMDICT_CREATE_SQL)
     with click.progressbar(data, width = 0, label = "writing jmdict") as bar:
       for e in bar:
-        c.execute("INSERT INTO entry VALUES (?,?,?,?,?,?)",
-                  (e.seq, e._rank(), str(e._freq()),
-                   e.isprio(), e.isnoun(), e.isverb()))
+        c.execute("INSERT INTO entry VALUES (?,?,?,?,?,?,?)",
+                  (e.seq, e.jlpt, e._rank(), str(e._freq()),
+                   e.prio_level(), e.isnoun(), e.isverb()))
         for k in e.kanji:
           c.execute("INSERT INTO kanji VALUES (?,?,?,?,?)",
                     (e.seq, k.elem, "".join(sorted(k.chars)),
                      "\n".join(k.info), k.prio))
-        for k in e.chars():
+        for k in sorted(e.chars()):
           c.execute("INSERT INTO kanji_code VALUES (?,?)",
                     (e.seq, ord(k)))
         for r in e.reading:
@@ -365,9 +499,10 @@ JMDICT_CREATE_SQL = """
 
   CREATE TABLE entry(
     seq INTEGER PRIMARY KEY ASC,
+    jlpt INTEGER,
     rank INTEGER,
     freq INTEGER,
-    prio BOOLEAN,
+    prio INTEGER,
     noun BOOLEAN,
     verb BOOLEAN
   );
@@ -376,7 +511,7 @@ JMDICT_CREATE_SQL = """
     elem TEXT,
     chars TEXT,
     info TEXT,
-    prio BOOLEAN,
+    prio INTEGER,
     FOREIGN KEY(entry) REFERENCES entry(seq)
   );
   CREATE TABLE kanji_code(
@@ -389,7 +524,7 @@ JMDICT_CREATE_SQL = """
     elem TEXT,
     restr TEXT,
     info TEXT,
-    prio BOOLEAN,
+    prio INTEGER,
     FOREIGN KEY(entry) REFERENCES entry(seq)
   );
   CREATE TABLE sense(
@@ -426,25 +561,27 @@ def setup(file = SQLITE_FILE):                                  # {{{1
   jmdict2sqldb(jmdict, file)
                                                                 # }}}1
 
-def nvp(noun, verb, prio):
-  if not any([noun, verb, prio]): return ""
+def search_filter(noun, verb, prio, jlpt):
+  if not any([noun, verb, prio, jlpt]): return ""
   s = []
   if noun and verb: s.append("(noun = 1 OR verb = 1)")
   elif noun       : s.append("noun = 1")
   elif verb       : s.append("verb = 1")
-  if prio         : s.append("prio = 1")
+  if prio         : s.append("prio >= {}".format(MINPRIO))
+  if jlpt         : s.append("({} <= jlpt AND jlpt <= {})"
+                             .format(*map(int, jlpt)))
   return "WHERE " + " AND ".join(s)
 
-def load_entry(c, seq):                                         # {{{1
+def load_entry(c, seq, jlpt):                                   # {{{1
   k = tuple(
       Kanji(r["elem"], frozenset(r["chars"]),
-            tuple(r["info"].splitlines()), bool(r["prio"]))
+            tuple(r["info"].splitlines()), r["prio"])
     for r in c.execute("SELECT * FROM kanji WHERE entry = ?" +
                        " ORDER BY rowid ASC", (seq,))
   )
   r = tuple(
       Reading(r["elem"], tuple(r["restr"].splitlines()),
-              tuple(r["info"].splitlines()), bool(r["prio"]))
+              tuple(r["info"].splitlines()), r["prio"])
     for r in c.execute("SELECT * FROM reading WHERE entry = ?" +
                        " ORDER BY rowid ASC", (seq,))
   )
@@ -456,51 +593,52 @@ def load_entry(c, seq):                                         # {{{1
     for r in c.execute("SELECT * FROM sense WHERE entry = ?" +
                        " ORDER BY rowid ASC", (seq,))
   )
-  return Entry(seq, *( tuple(x) for x in [k, r, s] ))
+  return Entry(seq, jlpt, *(k, r, s))
                                                                 # }}}1
 
 # TODO
 def search(q, langs = [LANGS[0]], max_results = None,           # {{{1
            noun = False, verb = False, prio = False,
-           file = SQLITE_FILE):
+           jlpt = None, file = SQLITE_FILE):
   fix_rank = lambda r: (r if r != F.NOFREQ else None)
   with sqlite_do(file) as c:
-    if q.lower() == "+random": q = "+#{}".format(random_seq(file))
+    if q.lower() == "+random":
+      q = "+#{}".format(random_seq(noun, verb, prio, jlpt, file))
     if re.fullmatch(r"\+#\s*\d+", q):
       seq = int(q[2:].strip())
-      for r in c.execute("SELECT rank FROM entry WHERE seq = ?", (seq,)):
-        yield load_entry(c, seq), fix_rank(r[0]) # #=1
+      for r in c.execute("SELECT rank, jlpt FROM entry WHERE seq = ?", (seq,)):
+        yield load_entry(c, seq, r[1]), fix_rank(r[0]) # #=1
     else:
       lang  = ",".join( "'" + l + "'" for l in langs if l in LANGS )
       limit = "LIMIT " + str(int(max_results)) if max_results else ""
-      fltr  = nvp(noun, verb, prio)
+      fltr  = search_filter(noun, verb, prio, jlpt)
+      ordr  = """ORDER BY IFNULL(prio, 0) >= {} DESC,
+                          rank = {} ASC, prio IS NULL ASC,
+                          rank ASC, jlpt DESC, prio DESC, seq ASC
+              """.format(MINPRIO, F.NOFREQ)                     # TODO
       if len(q) == 1 and M.iskanji(q):
         query = ("""
-          SELECT rank, seq FROM (
+          SELECT rank, seq, jlpt FROM (
             SELECT entry FROM kanji_code WHERE code = ?
           )
           INNER JOIN entry ON seq = entry
-          {}
-          ORDER BY prio DESC, rank ASC, seq ASC
-          {}
-        """.format(fltr, limit), (ord(q),))                   # safe!
+          {} {} {}
+        """.format(fltr, ordr, limit), (ord(q),))             # safe!
       elif M.iscjk(q):
         query = ("""
-          SELECT rank, seq FROM (
+          SELECT rank, seq, jlpt FROM (
               SELECT entry FROM kanji WHERE elem LIKE :q
             UNION
               SELECT entry FROM reading WHERE elem LIKE :q
           )
           INNER JOIN entry ON seq = entry
-          {}
-          ORDER BY prio DESC, rank ASC, seq ASC
-          {}
-        """.format(fltr, limit), dict(q = "%"+q+"%"))         # safe!
+          {} {} {}
+        """.format(fltr, ordr, limit), dict(q = "%"+q+"%"))   # safe!
       elif M.q2like(q):
         load_pcre_extension(c.connection)
         prms  = dict(q = M.q2like(q), re = M.q2rx(q))
         query = ("""
-          SELECT rank, seq FROM (
+          SELECT rank, seq, jlpt FROM (
               SELECT entry FROM kanji WHERE
                 elem LIKE :q AND elem REGEXP :re
             UNION
@@ -513,14 +651,12 @@ def search(q, langs = [LANGS[0]], max_results = None,           # {{{1
         """.format(lang)) + """
           )
           INNER JOIN entry ON seq = entry
-          {}
-          ORDER BY prio DESC, rank ASC, seq ASC
-          {}
-        """.format(fltr, limit), prms)                        # safe!
+          {} {} {}
+        """.format(fltr, ordr, limit), prms)                  # safe!
       else:
         load_pcre_extension(c.connection)
         query = ("""
-          SELECT rank, seq FROM (
+          SELECT rank, seq, jlpt FROM (
               SELECT entry FROM kanji WHERE elem REGEXP :re
             UNION
               SELECT entry FROM reading WHERE elem REGEXP :re
@@ -529,27 +665,38 @@ def search(q, langs = [LANGS[0]], max_results = None,           # {{{1
                 lang IN ({}) AND gloss REGEXP :re
           )
           INNER JOIN entry ON seq = entry
-          {}
-          ORDER BY prio DESC, rank ASC, seq ASC
-          {}
-        """.format(lang, fltr, limit), dict(re = M.q2rx(q)))  # safe!
-      for r, seq in [ tuple(r) for r in c.execute(*query) ]:  # eager!
-        yield load_entry(c, seq), fix_rank(r)
+          {} {} {}
+        """.format(lang, fltr, ordr, limit),                  # safe!
+          dict(re = M.q2rx(q)))
+      for r, s, j in [ tuple(r) for r in c.execute(*query) ]: # eager!
+        yield load_entry(c, s, j), fix_rank(r)
                                                                 # }}}1
 
 def by_freq(offset = 0, limit = 1000, file = SQLITE_FILE):
   with sqlite_do(file) as c:
-    ents = [ tuple(r) for r in c.execute("""
-        SELECT seq, rank FROM entry WHERE prio AND rank != {}
-          ORDER BY rank ASC LIMIT ? OFFSET ?
-        """.format(F.NOFREQ), (limit, offset)) ]              # safe!
-    for seq, rank in ents:
-      yield load_entry(c, seq), rank
+    q = """ SELECT seq, rank, jlpt FROM entry
+              WHERE prio >= {} AND rank != {}
+              ORDER BY rank ASC LIMIT ? OFFSET ?
+        """.format(MINPRIO, F.NOFREQ)                         # safe!
+    ents = [ tuple(r) for r in c.execute(q, (int(limit), int(offset))) ]
+    for seq, rank, jlpt in ents:
+      yield load_entry(c, seq, jlpt), rank
 
-def random_seq(file = SQLITE_FILE):
+def by_jlpt(n, offset = 0, limit = 1000, file = SQLITE_FILE):
   with sqlite_do(file) as c:
-    q = "SELECT seq FROM entry ORDER BY RANDOM() LIMIT 1"
-    return c.execute(q).fetchone()[0]
+    query =  (""" SELECT seq, jlpt FROM entry
+                    WHERE prio >= {} AND jlpt = ?
+                    ORDER BY rank ASC LIMIT ? OFFSET ?
+              """.format(MINPRIO), (int(n), int(limit), int(offset)))
+    for seq, jlpt in [ tuple(r) for r in c.execute(*query) ]:
+      yield load_entry(c, seq, jlpt)
+
+def random_seq(noun = False, verb = False, prio = False, jlpt = None,
+               file = SQLITE_FILE):
+  f = search_filter(noun, verb, prio, jlpt)
+  with sqlite_do(file) as c:
+    q = "SELECT seq FROM entry {} ORDER BY RANDOM() LIMIT 1".format(f)
+    return c.execute(q).fetchone()[0]                       # ^ safe!
 
 if __name__ == "__main__":
   if "--doctest" in sys.argv:
