@@ -5,10 +5,10 @@
 #
 # File        : jiten/kanji.py
 # Maintainer  : Felix C. Stegerman <flx@obfusk.net>
-# Date        : 2020-08-19
+# Date        : 2021-01-21
 #
-# Copyright   : Copyright (C) 2020  Felix C. Stegerman
-# Version     : v0.3.0
+# Copyright   : Copyright (C) 2021  Felix C. Stegerman
+# Version     : v0.3.5
 # License     : AGPLv3+
 #
 # --                                                            ; }}}1
@@ -29,23 +29,25 @@ KanjiDic.
 13108
 
 >>> [ x for x in kanjidic if x.char == "猫" ][0]
-Entry(char='猫', cat='KANJI', level='常用', strokes=11, freq=1702, jlpt=2, skip='1-3-8', rad=94, comp='⺨⽝⽥⾋犬犯猫田艸艹艾苗', var='貓', on=('ビョウ',), kun=('ねこ',), nanori=(), meaning=('cat',))
+Entry(char='猫', cat='KANJI', level='常用', strokes=11, freq=1702, jlpt=2, new_jlpt=3, skip='1-3-8', rad=94, comp='⺨⽝⽥⾋犬犯猫田艸艹艾苗', var='貓', on=('ビョウ',), kun=('ねこ',), nanori=(), meaning=('cat',))
 
 >>> [ x for x in kanjidic if x.char == "日" ][0]
-Entry(char='日', cat='KANJI', level='常用1', strokes=4, freq=1, jlpt=4, skip='3-3-1', rad=72, comp='⽇日', var='', on=('ニチ', 'ジツ'), kun=('ひ', '-び', '-か'), nanori=('あ', 'あき', 'いる', 'く', 'くさ', 'こう', 'す', 'たち', 'に', 'にっ', 'につ', 'へ'), meaning=('day', 'sun', 'Japan', 'counter for days'))
+Entry(char='日', cat='KANJI', level='常用1', strokes=4, freq=1, jlpt=4, new_jlpt=5, skip='3-3-1', rad=72, comp='⽇日', var='', on=('ニチ', 'ジツ'), kun=('ひ', '-び', '-か'), nanori=('あ', 'あき', 'いる', 'く', 'くさ', 'こう', 'す', 'たち', 'に', 'にっ', 'につ', 'へ'), meaning=('day', 'sun', 'Japan', 'counter for days'))
 
+>>> len([ x for x in kanjidic if x.level and x.level.startswith("常用") and x.level[-1].isdigit() ])
+1026
 >>> len([ x for x in kanjidic if x.level == "常用1" ])
 80
 >>> len([ x for x in kanjidic if x.level == "常用2" ])
 160
 >>> len([ x for x in kanjidic if x.level == "常用3" ])
 200
->>> len([ x for x in kanjidic if x.level == "常用4" ])
-197
+>>> len([ x for x in kanjidic if x.level == "常用4" ])  # 200 + 20 都道府県名
+220
 >>> len([ x for x in kanjidic if x.level == "常用5" ])
-197
+185
 >>> len([ x for x in kanjidic if x.level == "常用6" ])
-192
+181
 >>> len([ x for x in kanjidic if x.level == "常用" ])
 1110
 >>> len([ x for x in kanjidic if x.level == "人名" ])
@@ -59,7 +61,7 @@ Entry(char='日', cat='KANJI', level='常用1', strokes=4, freq=1, jlpt=4, skip=
 2501
 >>> len([ x for x in kanjidic if x.jlpt is not None ])
 2230
->>> len([ x for x in kanjidic if x.new_jlpt() is not None ])
+>>> len([ x for x in kanjidic if x.new_jlpt is not None ])
 2211
 >>> len([ x for x in kanjidic if x.skip is None ])
 952
@@ -96,12 +98,21 @@ True
 >>> all( ord(x[0]) - 0x2f00 == i for i, x in enumerate(RADICALS) )
 True
 
+>>> from .kana import kana2romaji
+>>> len([ kana2romaji(r) for e in kanjidic for r in e.on ])
+20994
+>>> len([ kana2romaji(r) for e in kanjidic for r in e.kun ])
+16024
+>>> len([ kana2romaji(r) for e in kanjidic for r in e.nanori ])
+3465
+
 """                                                             # }}}1
 
 import gzip, itertools, re, sys, unicodedata as UD
 import xml.etree.ElementTree as ET
 
 from collections import namedtuple
+from contextlib import contextmanager
 
 import click
 
@@ -115,13 +126,15 @@ KANJIVG_FILE    = M.resource_path("res/radicals/kanjivg.xml.gz")
 KRADFILE        = M.resource_path("res/radicals/kradfile.utf8")
 KRADFILE2       = M.resource_path("res/radicals/kradfile2.utf8")
 JLPT_FILE_BASE  = M.resource_path("res/jlpt/N")
+DATA_FILES      = (SQLITE_FILE, KANJIDIC_FILE, KANJIVG_FILE, KRADFILE,
+                   KRADFILE2)
 
 MAXE   = 25                                                     # TODO
 NOFREQ = 9999
 LEVELS = "常用1 常用2 常用3 常用4 常用5 常用6 常用 人名 人名(常用)".split()
 
 Entry = namedtuple("Entry", """
-  char cat level strokes freq jlpt skip rad comp var
+  char cat level strokes freq jlpt new_jlpt skip rad comp var
   on kun nanori meaning
 """.split())
 
@@ -135,7 +148,6 @@ Entry.canonical   = lambda e: canonical(e.char)
 Entry.radical     = lambda e: RADICALS[e.rad-1][1]
 Entry.name        = lambda e: UD.name(e.char)
 Entry.jmdict      = lambda e: J.search(e.char, max_results = MAXE)
-Entry.new_jlpt    = lambda e: JLPT.get(e.char)
 
 def canonical(c): return UD.normalize("NFC", c)
 
@@ -167,7 +179,7 @@ def variants(char, vs):
 
 def decode_variant(t, x):
   decode  = lambda b, x = "": b.decode("iso-2022-jp" + x)
-  dekuten = lambda: bytes( int(i) + 32 for i in x.split("-") )
+  dekuten = lambda: bytes( int(i) + 32 for i in x.split("-")[-2:] )
   if t == "jis208": return decode(b"\x1b$B"  + dekuten())
   if t == "jis212": return decode(b"\x1b$(D" + dekuten(), "-2")
   if t == "jis213": return decode(b"\x1b$(P" + dekuten(), "-2004")
@@ -206,10 +218,10 @@ def parse_kanjidic(kanjivg = None, file = KANJIDIC_FILE):       # {{{1
         meaning = tuple( m.text.strip() for m in e.findall(".//meaning")
                                         if m.get("m_lang") is None )
         if comp and not set(RADICALS[rad-1]).issubset(comp):
-          for x, y in "肉⽉ 白⽇ 曰⽇ 臼𦥑 匸⼕ 夊⼡ 夂久 人⼊ 入⼈".split():
+          for x, y in "肉⽉ 白⽇ 曰⽇ 臼𦥑 匸⼕ 夊⼡ 夂久 人⼊ 入⼈ 釆采".split():
             if x == RADICALS[rad-1][1] and y in comp: break
           else:
-            assert char in "為亀巨尭壮争単壷丗舅关"
+            assert char in "巨尭之冒丗关"
         comp = "".join(sorted(comp | set(RADICALS[rad-1] + char)))
         assert len(char) == 1
         assert 1 <= rad <= 214
@@ -220,8 +232,9 @@ def parse_kanjidic(kanjivg = None, file = KANJIDIC_FILE):       # {{{1
         assert all( "\n" not in x for x in kun )
         assert all( "\n" not in x for x in nanori )
         assert all( "\n" not in x for x in meaning )
-        data.append(Entry(char, category(char), lvl, strokes, freq, jlpt,
-                          skip, rad, comp, var, on, kun, nanori, meaning))
+        data.append(Entry(char, category(char), lvl, strokes, freq,
+                          jlpt, JLPT.get(char), skip, rad, comp, var,
+                          on, kun, nanori, meaning))
       return data
                                                                 # }}}1
 
@@ -264,7 +277,7 @@ def parse_kanjivg(file = KANJIVG_FILE, kradfile = KRADFILE,     # {{{1
 def load_jlpt(base = JLPT_FILE_BASE):
   data = {}
   for level in "12345":
-    with open(base + level) as f:
+    with open(base + level + "-kanji") as f:
       for c in f.readline().strip():
         assert M.iskanji(c)
         assert c not in data
@@ -281,8 +294,8 @@ def kanjidic2sqldb(data, file = SQLITE_FILE):                   # {{{1
         c.execute("INSERT INTO entry VALUES ({})"
                   .format(",".join("?"*(len(Entry._fields)+1))),
                   (ord(e.char), e.char, e.cat, e.level, e.strokes,
-                   e.freq, e.jlpt, e.skip, e.rad, e.comp, e.var,
-                   "\n".join(e.on), "\n".join(e.kun),
+                   e.freq, e.jlpt, e.new_jlpt, e.skip, e.rad, e.comp,
+                   e.var, "\n".join(e.on), "\n".join(e.kun),
                    "\n".join(e.nanori), "\n".join(e.meaning)))
         for k in e.comp:
           c.execute("INSERT INTO comp VALUES(?,?)", (ord(e.char), ord(k)))
@@ -301,6 +314,7 @@ KANJIDIC_CREATE_SQL = """
     strokes INTEGER,
     freq INTEGER,
     jlpt INTEGER,
+    new_jlpt INTEGER,
     skip TEXT,
     rad INTEGER,
     comp TEXT,
@@ -380,24 +394,40 @@ def search(q, max_results = None, file = SQLITE_FILE):          # {{{1
 def by_freq(file = SQLITE_FILE):
   with sqlite_do(file) as c:
     for r in c.execute("""
-        SELECT char, freq FROM entry
+        SELECT char, freq, on_, kun, meaning FROM entry
           WHERE freq IS NOT NULL ORDER BY freq ASC
         """):
-      yield r["char"], r["freq"]
+      yield (r["char"], r["freq"]) + _readmean(r)
 
 def by_level(level, file = SQLITE_FILE):
   with sqlite_do(file) as c:
     for r in c.execute("""
-        SELECT char FROM entry WHERE level = ? ORDER BY code ASC
+        SELECT char, on_, kun, meaning FROM entry
+          WHERE level = ? ORDER BY code ASC
         """, (level,)):
-      yield r["char"]
+      yield (r["char"],) + _readmean(r)
 
-def by_jlpt():
+def by_jlpt(file = SQLITE_FILE):
   data = { int(l): [] for l in "12345" }
-  for char, level in JLPT.items():
-    data[level].append(char)
+  with sqlite_do(file) as c:
+    for char, level in JLPT.items():
+      data[level].append((char,) + readmean(char, c))
   for level in "54321":
-    yield int(level), "".join(sorted(data[int(level)]))
+    yield int(level), tuple(sorted(data[int(level)]))
+
+@contextmanager
+def readmeans(file = SQLITE_FILE):
+  with sqlite_do(file) as c:
+    yield lambda char: readmean(char, c)
+
+def readmean(char, c):
+  r = c.execute("SELECT on_, kun, meaning FROM entry WHERE code = ?",
+                (ord(char),)).fetchone()
+  return _readmean(r) if r else None
+
+def _readmean(r):
+  return (tuple(r["on_"].splitlines() + r["kun"].splitlines()),
+          tuple(r["meaning"].splitlines()))
 
 def random(file = SQLITE_FILE):
   with sqlite_do(file) as c:
