@@ -5,10 +5,10 @@
 #
 # File        : jiten/pitch.py
 # Maintainer  : Felix C. Stegerman <flx@obfusk.net>
-# Date        : 2020-08-30
+# Date        : 2021-01-19
 #
-# Copyright   : Copyright (C) 2020  Felix C. Stegerman
-# Version     : v0.3.2
+# Copyright   : Copyright (C) 2021  Felix C. Stegerman
+# Version     : v0.3.5
 # License     : AGPLv3+
 #
 # --                                                            ; }}}1
@@ -19,8 +19,8 @@ r"""
 Pitch Accent from Wadoku.
 
 >>> pitch = parse_pitch()
->>> len(pitch)
-141798
+>>> len(pitch) + len(BLACKLIST)
+205092
 
 >>> [ x for x in pitch if x[0] == "小猫" ][0]
 ['小猫', 'こねこ', '2']
@@ -41,9 +41,21 @@ Pitch Accent from Wadoku.
 >>> with_accent("はし", 2)
 'はꜛしꜜ'
 
+>>> r = [ x for x in pitch if x[0] == "以心伝心" ][0]
+>>> r
+['以心伝心', 'いしん—でんしん', '1']
+>>> with_pitch(dict(reading = r[1], accent = r[2]))
+'いꜜしん･でんしん'
+
+>>> r = [ x for x in pitch if x[0] == "精一杯" ][0]
+>>> r
+['精一杯', 'せい—いっぱい', '3']
+>>> with_pitch(dict(reading = r[1], accent = r[2]))
+'せꜛい･いꜜっぱい'
+
 """                                                             # }}}1
 
-import re, sys
+import functools, os, re, sys
 
 import click
 
@@ -52,8 +64,14 @@ from .sql import sqlite_do
 
 SQLITE_FILE = M.resource_path("res/pitch.sqlite3")
 PITCH_FILE  = M.resource_path("res/pitch/PITCH")
+DATA_FILES  = (SQLITE_FILE, PITCH_FILE)
 
-MORASPLIT   = re.compile(r"(.[ぁぃぅぇぉゃょゅァィゥェォャュョ]?)")
+# NB: skip ･ for e.g. せꜛい･いꜜっぱい
+MORASPLIT   = re.compile(r"(･?.[ぁぃぅぇぉゃょゅァィゥェォャュョ]?)")
+
+# TODO
+BLACKLIST   = """ぎっこんばっこん ぎっこんばったん ぎったんばったん
+                 電子マネー対応自販機 電子マネー自販機""".split()
 
 # TODO
 def with_accent(text, pos):
@@ -71,7 +89,10 @@ def parse_pitch(file = PITCH_FILE):
   with open(file) as f:
     with click.progressbar(f, width = 0, label = "reading pitch") as bar:
       for line in bar:
-        data.append(line.split())
+        kanji, reading, accent = row = line.split()
+        if kanji in BLACKLIST: continue
+        assert len(reading.split("—")) >= len(accent.split("—"))
+        data.append(row)
   return data
 
 def pitch2sqldb(data, file = SQLITE_FILE):
@@ -99,14 +120,24 @@ def setup(file = SQLITE_FILE):
 # TODO
 def get_pitch(reading, kanjis, file = SQLITE_FILE):
   rd = reading.replace("・", "")
-  with sqlite_do(file) as c:
-    for k in ( k.replace("・", "") for k in kanjis ):
-      for r in c.execute("SELECT * FROM entry WHERE kanji = ?", (k,)):
-        if r["reading"].replace("—", "") != rd: continue
-        ra = zip(r["reading"].split("—"), r["accent"].split("—"))
-        return "･".join( with_accent(r, int(a)) for r, a in ra )
+  if have_pitch(file):
+    with sqlite_do(file) as c:
+      for k in ( k.replace("・", "") for k in kanjis ):
+        for r in c.execute("SELECT * FROM entry WHERE kanji = ?", (k,)):
+          if r["reading"].replace("—", "") != rd: continue
+          return with_pitch(r)
   return None
 
+def with_pitch(r):
+  sr, sa  = r["reading"].split("—"), r["accent"].split("—")
+  rs      = sr[:len(sa)-1] + ["･".join(sr[len(sa)-1:])]
+  return "･".join( with_accent(r, a if a is None else int(a))
+                   for r, a in zip(rs, sa) )
+
+@functools.lru_cache(maxsize = None)
+def have_pitch(file = SQLITE_FILE):
+  return os.path.exists(file)
+
 if __name__ == "__main__":
   if "--doctest" in sys.argv:
     verbose = "--verbose" in sys.argv
